@@ -4,6 +4,7 @@ import datetime as dt
 import gzip
 import json
 import os
+import re
 import shutil
 import sys
 import time
@@ -34,6 +35,7 @@ from .changesets import ChangesetToolKit
 users_temp = {}
 users = {}
 hashtag_changesets = {}
+countries_changesets = {}
 
 # read the GeoJSON file
 countries_df = gpd.read_file("data/countries_un.geojson")
@@ -43,9 +45,9 @@ def collect_changefile_stats(user, uname, changeset, version, tags, osm_type):
     tags_to_collect = list(additional_tags) if additional_tags else None
     if version == 1:
         action = "create"
-    elif version > 1:
+    if version > 1:
         action = "modify"
-    elif version == 0:
+    if version == 0:
         action = "delete"
     if user in users:
         users[user]["name"] = uname
@@ -57,6 +59,7 @@ def collect_changefile_stats(user, uname, changeset, version, tags, osm_type):
             for ch in hashtag_changesets[changeset]:
                 if ch not in users[user]["countries"]:
                     users[user]["countries"].append(ch)
+
         users[user][osm_type][action] += 1
         if osm_type == "nodes" and tags:
             users[user]["poi"][action] += 1
@@ -71,6 +74,10 @@ def collect_changefile_stats(user, uname, changeset, version, tags, osm_type):
             for tag in tags_to_collect:
                 if tag in tags:
                     users[user][tag][action] += 1
+        if country:
+            for ch in countries_changesets[changeset]:
+                if ch not in users[user]["hashtags"]:
+                    users[user]["hashtags"].append(ch)
 
     else:
         users[user] = {
@@ -80,13 +87,14 @@ def collect_changefile_stats(user, uname, changeset, version, tags, osm_type):
             "nodes": {"create": 0, "modify": 0, "delete": 0},
             "ways": {"create": 0, "modify": 0, "delete": 0},
             "relations": {"create": 0, "modify": 0, "delete": 0},
-            "poi": {"create": 0, "modify": 0, "delete": 0},  # nodes that has tags
+            "poi": {"create": 0, "modify": 0},  # nodes that has tags
         }
         if hashtags:
             users[user]["countries"] = hashtag_changesets[changeset]
+
         if tags_to_collect:
             for tag in tags_to_collect:
-                users[user][tag] = {"create": 0, "modify": 0, "delete": 0}
+                users[user][tag] = {"create": 0, "modify": 0}
         users_temp[user] = {"changesets": []}
         if changeset not in users_temp[user]["changesets"]:
             users_temp[user]["changesets"].append(changeset)
@@ -95,7 +103,6 @@ def collect_changefile_stats(user, uname, changeset, version, tags, osm_type):
         if wild_tags:
             users[user]["tags_create"] = {}
             users[user]["tags_modify"] = {}
-            users[user]["tags_delete"] = {}
 
             for tag, value in tags:
                 users[user][f"tags_{action}"][tag] = 1
@@ -103,10 +110,12 @@ def collect_changefile_stats(user, uname, changeset, version, tags, osm_type):
             for tag in tags_to_collect:
                 if tag in tags:
                     users[user][tag][action] = 1
+        if country:
+            users[user]["hashtags"] = countries_changesets[changeset]
 
 
 def calculate_stats(user, uname, changeset, version, tags, osm_type):
-    if hashtags:  # intersect with changesets
+    if (hashtags and country) or hashtags:  # intersect with changesets
         if (
             len(hashtag_changesets) > 0
         ):  # make sure there are changesets to intersect if not meaning hashtag changeset not found no need to go for changefiles
@@ -114,6 +123,12 @@ def calculate_stats(user, uname, changeset, version, tags, osm_type):
                 collect_changefile_stats(
                     user, uname, changeset, version, tags, osm_type
                 )
+    elif country:
+        if len(countries_changesets) > 0:
+            if changeset in countries_changesets.keys():
+                collect_changefile_stats(
+                    user, uname, changeset, version, tags, osm_type
+                )
     else:  # collect everything
         collect_changefile_stats(user, uname, changeset, version, tags, osm_type)
 
@@ -123,10 +138,32 @@ class ChangesetHandler(osmium.SimpleHandler):
         super(ChangesetHandler, self).__init__()
 
     def changeset(self, c):
-        if "comment" in c.tags:
-            if any(elem.lower() in c.tags["comment"].lower() for elem in hashtags):
-                hashtag_changesets[c.id] = []
-                # get bbox
+        country_check = False
+        if hashtags:
+            if "comment" in c.tags:
+                if any(elem.lower() in c.tags["comment"].lower() for elem in hashtags):
+                    if c.id not in hashtag_changesets.keys():
+                        hashtag_changesets[c.id] = []
+                    # get bbox
+                    bounds = str(c.bounds)
+                    bbox_list = bounds.strip("()").split(" ")
+                    minx, miny = bbox_list[0].split("/")
+                    maxx, maxy = bbox_list[1].split("/")
+                    bbox = box(float(minx), float(miny), float(maxx), float(maxy))
+                    # Create a point for the centroid of the bounding box
+                    centroid = bbox.centroid
+                    intersected_rows = countries_df[countries_df.intersects(centroid)]
+                    for i, row in intersected_rows.iterrows():
+                        if row["name"] not in hashtag_changesets[c.id]:
+                            if country:
+                                country_check = True
+                                if row["name"] == country:
+                                    hashtag_changesets[c.id].append(row["name"])
+                            else:
+                                hashtag_changesets[c.id].append(row["name"])
+
+        if not country_check:  # hash tag not supplied
+            if country:
                 bounds = str(c.bounds)
                 bbox_list = bounds.strip("()").split(" ")
                 minx, miny = bbox_list[0].split("/")
@@ -136,8 +173,14 @@ class ChangesetHandler(osmium.SimpleHandler):
                 centroid = bbox.centroid
                 intersected_rows = countries_df[countries_df.intersects(centroid)]
                 for i, row in intersected_rows.iterrows():
-                    if row["name"] not in hashtag_changesets[c.id]:
-                        hashtag_changesets[c.id].append(row["name"])
+                    if row["name"] == country:
+                        if c.id not in countries_changesets.keys():
+                            countries_changesets[c.id] = []
+                        if "comment" in c.tags:
+                            hashtags_comment = re.findall(r"#[\w-]+", c.tags["comment"])
+                            for tag in hashtags_comment:
+                                if tag not in countries_changesets[c.id]:
+                                    countries_changesets[c.id].append(tag)
 
 
 class ChangefileHandler(osmium.SimpleHandler):
@@ -146,16 +189,24 @@ class ChangefileHandler(osmium.SimpleHandler):
 
     def node(self, n):
         if n.timestamp >= start_date_utc and n.timestamp <= end_date_utc:
-            calculate_stats(n.uid, n.user, n.changeset, n.version, n.tags, "nodes")
+            version = n.version
+            if n.deleted:
+                version = 0
+            calculate_stats(n.uid, n.user, n.changeset, version, n.tags, "nodes")
 
     def way(self, w):
         if w.timestamp >= start_date_utc and w.timestamp <= end_date_utc:
-
-            calculate_stats(w.uid, w.user, w.changeset, w.version, w.tags, "ways")
+            version = w.version
+            if w.deleted:
+                version = 0
+            calculate_stats(w.uid, w.user, w.changeset, version, w.tags, "ways")
 
     def relation(self, r):
         if r.timestamp >= start_date_utc and r.timestamp <= end_date_utc:
-            calculate_stats(r.uid, r.user, r.changeset, r.version, r.tags, "relations")
+            version = r.version
+            if r.deleted:
+                version = 0
+            calculate_stats(r.uid, r.user, r.changeset, version, r.tags, "relations")
 
 
 def process_changefiles(url):
@@ -228,6 +279,7 @@ def process_changesets(url):
         shutil.copyfileobj(f_in, f_out)
 
     handler.apply_file(file_path[:-3])
+    print(f"Finished {url}")
 
 
 def get_download_urls_changefiles(
@@ -314,8 +366,15 @@ def auth(username, password):
 
 def main():
     parser = argparse.ArgumentParser()
-    parser.add_argument("--start_date", help="Start date in the format YYYY-MM-DD")
-    parser.add_argument("--end_date", help="End date in the format YYYY-MM-DD")
+    parser.add_argument(
+        "--start_date",
+        help="Start date in the format YYYY-MM-DD HH:M:Sz eg: 2023-01-28 17:43:09+05:45",
+    )
+    parser.add_argument(
+        "--end_date",
+        help="End date in the format YYYY-MM-DD HH:M:Sz eg:2023-01-28 17:43:09+05:45",
+        default=dt.datetime.now(),
+    )
     parser.add_argument(
         "--username",
         default=None,
@@ -338,6 +397,12 @@ def main():
         default="stats",
         help="Output stat file name",
     )
+    parser.add_argument(
+        "--country",
+        default=None,
+        help="Country name to extract (get name from data/un_countries) : Only viable until day stats since changeset replication is available for minute, avoid using for geofabrik url since geofabrik already gives country level changefiles",
+    )
+
     parser.add_argument(
         "--tags",
         nargs="+",
@@ -361,12 +426,20 @@ def main():
     parser.add_argument(
         "--rows",
         type=int,
-        help="No fo top rows to extract , to extract top 100 , pass 100",
+        help="No. of top rows to extract , to extract top 100 , pass 100",
+    )
+    
+    parser.add_argument(
+        "--workers",
+        type=int,
+        default=None,
+        help="No. of Parallel workers to assign : Default is no of cpu available , Be aware to use this max no of workers may cause overuse of resources",
     )
 
     parser.add_argument(
-        "--url", required=True, help="Your public Geofabrik Download URL "
+        "--url", default="https://planet.openstreetmap.org/replication/minute", help="Your public OSM Change Replication URL "
     )
+
     parser.add_argument("--extract_last_week", action="store_true", default=False)
     parser.add_argument("--extract_last_day", action="store_true", default=False)
     parser.add_argument("--extract_last_month", action="store_true", default=False)
@@ -405,7 +478,7 @@ def main():
             dt.datetime.strptime(args.start_date, "%Y-%m-%d %H:%M:%S%z"), args.timezone
         )
 
-    if not args.start_date or not args.end_date:
+    if not args.start_date:
         if (
             args.extract_last_week
             or args.extract_last_day
@@ -415,22 +488,30 @@ def main():
         ):
             pass
         else:
-            print("ERR: Supply start_date & End_date")
+            print("ERR: Supply start_date")
             sys.exit()
 
     if args.end_date:
-        end_date = strip_utc(
-            dt.datetime.strptime(args.end_date, "%Y-%m-%d %H:%M:%S%z"), args.timezone
-        )
+        end_date = args.end_date
+        if not isinstance(end_date, datetime):
+            end_date = dt.datetime.strptime(args.end_date, "%Y-%m-%d %H:%M:%S%z")
+
+        end_date = strip_utc(end_date, args.timezone)
+    if args.country:
+        if not countries_df["name"].isin([args.country]).any():
+            print("Country doesn't exists : Refer to data/countries_un.csv name column")
+            sys.exit()
     start_time = time.time()
 
     global additional_tags
     global cookies
     global wild_tags
     global hashtags
+    global country
     wild_tags = args.wild_tags
     additional_tags = args.tags
     hashtags = args.hashtags
+    country = args.country
     cookies = None
     if "geofabrik" in args.url.lower():
         if args.username is None:
@@ -479,14 +560,14 @@ def main():
         sys.exit()
 
     if (end_date - start_date).days > 1:
-        if args.hashtags:
+        if args.hashtags or args.country:
             print(
-                "Replication for Changeset is minutely , To process more than a day data it might take a while"
+                "Warning : Replication for Changeset is minutely , To process more than a day data it might take a while , Use --force to ignore this warning"
             )
             if not args.force:
                 sys.exit()
 
-    if args.hashtags:
+    if args.hashtags or args.country:
 
         Changeset = ChangesetToolKit()
         (
@@ -497,9 +578,14 @@ def main():
         print(
             f"You have supplied start_date as : {start_date} and end_date as : {end_date} , Processing Changeset from {strip_utc(Changeset.sequence_to_timestamp(changeset_start_seq),args.timezone)} to {strip_utc(Changeset.sequence_to_timestamp(changeset_end_seq),args.timezone)}"
         )
-        with concurrent.futures.ThreadPoolExecutor() as executor:
+
+        max_workers = os.cpu_count() if not args.workers else args.workers 
+ 
+        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
             # Use `map` to apply the `download_image` function to each element in the `urls` list
             executor.map(process_changesets, changeset_download_urls)
+            # executor.shutdown(wait=True)
+
         print("Changeset Processing Finished")
         end_date = strip_utc(
             Changeset.sequence_to_timestamp(changeset_end_seq), args.timezone
@@ -516,7 +602,7 @@ def main():
     ) = get_download_urls_changefiles(start_date, end_date, args.url, args.timezone)
     if server_ts < end_date:
         print(
-            "Warning : End date data is not available at server, Changing to latest available date "
+            f"Warning : End date data is not available at server, Changing to latest available date {server_ts}"
         )
         end_date = server_ts
         if start_date >= server_ts:
@@ -525,20 +611,22 @@ def main():
     global end_date_utc
     global start_date_utc
 
-    start_date_utc = start_date
-    end_date_utc = end_date
+    start_date_utc = start_date.astimezone(dt.timezone.utc)
+    end_date_utc = end_date.astimezone(dt.timezone.utc)
     temp_path = os.path.join(os.getcwd(), "temp/changefiles")
     if not os.path.exists(temp_path):
         os.makedirs(temp_path)
 
     # Use the ThreadPoolExecutor to download the images in parallel
-    with concurrent.futures.ThreadPoolExecutor() as executor:
+    with concurrent.futures.ThreadPoolExecutor(max_workers=os.cpu_count() if not args.workers else args.workers) as executor:
         # Use `map` to apply the `download_image` function to each element in the `urls` list
         executor.map(process_changefiles, download_urls)
+
+        # executor.shutdown(wait=True)
     print("Changefiles Processing Finished")
     os.chdir(os.getcwd())
     shutil.rmtree("temp")
-    if len(users) > 1:
+    if len(users) >= 1:
         # print(users)
         if args.wild_tags:
             for user in users:
@@ -560,15 +648,6 @@ def main():
                         )
                     )
                 )
-                users[user]["tags_delete"] = json.dumps(
-                    dict(
-                        sorted(
-                            users[user]["tags_delete"].items(),
-                            key=lambda item: item[1],
-                            reverse=True,
-                        )
-                    )
-                )
         df = pd.json_normalize(list(users.values()))
 
         if hashtags:
@@ -596,7 +675,7 @@ def main():
             # Get the column names of the DataFrame
             cols = df.columns.tolist()
             # Identify the column names that you want to move
-            cols_to_move = ["tags_create", "tags_modify", "tags_delete"]
+            cols_to_move = ["tags_create", "tags_modify"]
             # Remove the columns to move from the list of column names
             cols = [col for col in cols if col not in cols_to_move]
             # Add the columns to move to the end of the list of column names
@@ -604,6 +683,11 @@ def main():
             # Reindex the DataFrame with the new order of column names
             df = df.reindex(columns=cols)
 
+        if country:
+            df["hashtags"] = df["hashtags"].apply(lambda x: ",".join(map(str, x)))
+            column_to_move = "hashtags"
+            df = df.assign(**{column_to_move: df.pop(column_to_move)})
+
         start_date = in_local_timezone(start_date, args.timezone)
         end_date = in_local_timezone(end_date, args.timezone)
 
@@ -618,7 +702,7 @@ def main():
                 else df.head(100)
             )  # 100 as max rows for image format
             dfi.export(
-                df_img.drop(columns=["tags_create", "tags_modify", "tags_delete"])
+                df_img.drop(columns=["tags_create", "tags_modify"])
                 if args.wild_tags
                 else df_img,
                 f"{fname}.png",
